'''
Cheetah is an open source template engine and code generation tool.

It can be used standalone or combined with other tools and frameworks. Web
development is its principle use, but Cheetah is very flexible and is also being
used to generate C++ game code, Java, sql, form emails and even Python code.

Homepage
    http://www.cheetahtemplate.org/

Documentation
    http://cheetahtemplate.org/learn.html

Mailing list
cheetahtemplate-discuss@lists.sourceforge.net
Subscribe at 
    http://lists.sourceforge.net/lists/listinfo/cheetahtemplate-discuss
'''

from Version import *
